monitor.o monitor.d : monitor.c genesis/sbcl.h lispobj.h runtime.h align.h parse.h \
 vars.h print.h genesis/vector.h lispobj.h arch.h os.h target-os.h \
 target-arch-os.h target-arch.h thread.h globals.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 interrupt.h validate.h gc-typedefs.h interr.h search.h lispregs.h \
 target-lispregs.h genesis/regnames.h genesis/instance.h instance.inc \
 genesis/primitive-objects.h genesis/number-types.h genesis/closure.h \
 genesis/cons.h genesis/instance.h genesis/return-pc.h genesis/sap.h \
 genesis/symbol.h genesis/thread.h genesis/value-cell.h genesis/vector.h \
 genesis/weak-pointer.h genesis/gc-tables.h genesis/unwind-block.h \
 genesis/hash-table.h gc.h gc-assert.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc ../../tlsf-bsd/tlsf/tlsf.h
