runtime.o runtime.d : runtime.c genesis/sbcl.h runtime.h lispobj.h align.h globals.h \
 os.h target-os.h target-arch-os.h target-arch.h interr.h interrupt.h \
 genesis/static-symbols.h thread.h genesis/symbol.h lispobj.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h genesis/vector.h validate.h \
 gc-typedefs.h arch.h gc.h gc-assert.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc core.h lispregs.h target-lispregs.h \
 genesis/regnames.h
