coreparse.o coreparse.d : coreparse.c genesis/sbcl.h os.h runtime.h lispobj.h align.h \
 target-os.h target-arch-os.h target-arch.h globals.h core.h arch.h \
 thread.h genesis/symbol.h lispobj.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h genesis/vector.h interrupt.h \
 validate.h gc-typedefs.h interr.h gc.h gc-assert.h immobile-space.h \
 gencgc-impl.h gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc graphvisit.h genesis/instance.h instance.inc \
 ../../output/build-id.inc genesis/gc-tables.h genesis/hash-table.h \
 var-io.h genesis/split-ordered-list.h
