/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_TAGNAMES
#define SBCL_GENESIS_TAGNAMES
#include <stddef.h>

static const char *lowtag_names[] = {
    "even fixnum",
    "instance pointer",
    "other immediate 0",
    "list pointer",
    "odd fixnum",
    "fun pointer",
    "other immediate 1",
    "other pointer"
};

const char *widetag_names[] = {
    "unknown [0]",
    "unknown [1]",
    "bignum",
    "ratio",
    "sfloat",
    "dfloat",
    "cplxnum",
    "cplx-sfloat",
    "cplx-dfloat",
    "symbol",
    "sap",
    "codeblob",
    "instance",
    "funinstance",
    "simplefun",
    "closure",
    "unknown [16]",
    "value-cell",
    "char",
    "unbound-marker",
    "weakptr",
    "fdefn",
    "unknown [22]",
    "unknown [23]",
    "unknown [24]",
    "filler",
    "unknown [26]",
    "unknown [27]",
    "unknown [28]",
    "unknown [29]",
    "unknown [30]",
    "unknown [31]",
    "simple-array",
    "simple-array-NIL",
    "simple-vec",
    "simple-bit-vec",
    "UB2-vec",
    "UB4-vec",
    "UB7-vec",
    "UB8-vec",
    "UB15-vec",
    "UB16-vec",
    "Ufixnum-vec",
    "UB31-vec",
    "UB32-vec",
    "SB8-vec",
    "SB16-vec",
    "fixnum-vec",
    "SB32-vec",
    "sfloat-vec",
    "dfloat-vec",
    "cplx-sfloat-vec",
    "cplx-dfloat-vec",
    "simple-base-str",
    "simple-char-str",
    "fancy-base-str",
    "fancy-char-str",
    "fancy-bit-vec",
    "fancy-vec",
    "fancy-array",
    "unknown [60]",
    "unknown [61]",
    "unknown [62]",
    "unknown [63]"
};
static char *symbol_slots[] = {
 "fdefn: ", "value: ", "info: ", "name: ", "hash: ", "package_id: ", NULL
};
static char *ratio_slots[] = {
 "numerator: ", "denominator: ", NULL
};
static char *complex_slots[] = {
 "real: ", "imag: ", NULL
};
static char *code_slots[] = {
 "boxed_size: ", "fixups: ", "debug_info: ", NULL
};
static char *simple_fun_slots[] = {
 "self: ", NULL
};
static char *closure_slots[] = {
 "fun: ", NULL
};
static char *funcallable_instance_slots[] = {
 "trampoline: ", "function: ", "layout: ", NULL
};
static char *weak_pointer_slots[] = {
 "value: ", "next: ", NULL
};
static char *fdefn_slots[] = {
 "name: ", "fun: ", "raw_addr: ", NULL
};
static char *value_cell_slots[] = {
 "value: ", NULL
};
#endif
