gc-common.o gc-common.d : gc-common.c genesis/sbcl.h runtime.h lispobj.h align.h os.h \
 target-os.h target-arch-os.h target-arch.h interr.h globals.h \
 interrupt.h genesis/static-symbols.h thread.h genesis/symbol.h lispobj.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h genesis/vector.h validate.h \
 gc-typedefs.h lispregs.h target-lispregs.h genesis/regnames.h arch.h \
 gc.h gc-assert.h immobile-space.h gencgc-impl.h gencgc-alloc-region.h \
 genesis/cardmarks.h genesis/cons.h genesis/weak-pointer.h hopscotch.h \
 code.h genesis/closure.h forwarding-ptr.inc genesis/primitive-objects.h \
 genesis/number-types.h genesis/closure.h genesis/cons.h \
 genesis/instance.h instance.inc genesis/return-pc.h genesis/sap.h \
 genesis/symbol.h genesis/thread.h genesis/value-cell.h genesis/vector.h \
 genesis/weak-pointer.h genesis/hash-table.h genesis/split-ordered-list.h \
 genesis/compiled-debug-info.h var-io.h search.h murmur_hash.h \
 incremental-compact.h weak-hash-pred.inc genesis/gc-tables.h
