thread.o thread.d : thread.c genesis/sbcl.h runtime.h lispobj.h align.h validate.h \
 thread.h globals.h os.h target-os.h target-arch-os.h target-arch.h \
 genesis/symbol.h lispobj.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/static-symbols.h genesis/thread.h \
 gencgc-alloc-region.h genesis/vector.h interrupt.h gc-typedefs.h arch.h \
 genesis/cons.h genesis/instance.h instance.inc interr.h gc-assert.h gc.h \
 immobile-space.h gencgc-impl.h gencgc-alloc-region.h genesis/cardmarks.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc pseudo-atomic.h lispregs.h target-lispregs.h \
 genesis/regnames.h atomiclog.inc
