/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_STATIC_SYMBOLS
#define SBCL_GENESIS_STATIC_SYMBOLS
#define NIL LISPOBJ(0x4100033)
#define LISP_T LISPOBJ(0x410004F)
#define FREE_INTERRUPT_CONTEXT_INDEX LISPOBJ(0x410006F)
#define ALLOW_WITH_INTERRUPTS LISPOBJ(0x410008F)
#define UNBLOCK_DEFERRABLES_ON_ENABLING_INTERRUPTS_P LISPOBJ(0x41000AF)
#define INTERRUPTS_ENABLED LISPOBJ(0x41000CF)
#define INTERRUPT_PENDING LISPOBJ(0x41000EF)
#define IN_WITHOUT_GCING LISPOBJ(0x410010F)
#define GC_INHIBIT LISPOBJ(0x410012F)
#define GC_PENDING LISPOBJ(0x410014F)
#define UNWEAKENED_VECTORS LISPOBJ(0x410016F)
#define PINNED_OBJECTS LISPOBJ(0x410018F)
#define GC_PIN_CODE_PAGES LISPOBJ(0x41001AF)
#define CURRENT_CATCH_BLOCK LISPOBJ(0x41001CF)
#define CURRENT_UNWIND_PROTECT_BLOCK LISPOBJ(0x41001EF)
#define IMMOBILE_CODEBLOB_TREE LISPOBJ(0x410020F)
#define IMMOBILE_CODEBLOB_VECTOR LISPOBJ(0x410022F)
#define DYNSPACE_CODEBLOB_TREE LISPOBJ(0x410024F)
#define LINKAGE_NAME_MAP LISPOBJ(0x410026F)
#define FINALIZER_STORE LISPOBJ(0x410028F)
#define FINALIZER_REHASHLIST LISPOBJ(0x41002AF)
#define FINALIZERS_TRIGGERED LISPOBJ(0x41002CF)
#define RUN_GC_HOOKS LISPOBJ(0x41002EF)
#define BINDING_STACK_START LISPOBJ(0x410030F)
#define CONTROL_STACK_START LISPOBJ(0x410032F)
#define CONTROL_STACK_END LISPOBJ(0x410034F)
#define STEPPING LISPOBJ(0x410036F)
#define FREE_TLS_INDEX LISPOBJ(0x410038F)
#define TLS_SYMBOL_MAP LISPOBJ(0x41003AF)
#define GC_EPOCH LISPOBJ(0x41003CF)
#define DATA_VECTOR_REFFERS LISPOBJ(0x41003EF)
#define DATA_VECTOR_REFFERS_CHECK_BOUNDS LISPOBJ(0x410040F)
#define DATA_VECTOR_SETTERS LISPOBJ(0x410042F)
#define DATA_VECTOR_SETTERS_CHECK_BOUNDS LISPOBJ(0x410044F)
#define BINDING_STACK_POINTER LISPOBJ(0x410046F)
#define PSEUDO_ATOMIC_ATOMIC LISPOBJ(0x410048F)
#define PSEUDO_ATOMIC_INTERRUPTED LISPOBJ(0x41004AF)
#define LFLIST_TAIL_ATOM LISPOBJ(0x41004C9)
#define INTERFACE_SYMBOLS ((lispobj*)(1645 + NIL))
#define SUB_GC_fname_index 0
#define POST_GC_fname_index 1
#define INTERNAL_ERROR_fname_index 2
#define CONTROL_STACK_EXHAUSTED_ERROR_fname_index 3
#define BINDING_STACK_EXHAUSTED_ERROR_fname_index 4
#define ALIEN_STACK_EXHAUSTED_ERROR_fname_index 5
#define HEAP_EXHAUSTED_ERROR_fname_index 6
#define UNDEFINED_ALIEN_VARIABLE_ERROR_fname_index 7
#define MEMORY_FAULT_ERROR_fname_index 8
#define UNHANDLED_TRAP_ERROR_fname_index 9
#define HANDLE_BREAKPOINT_fname_index 10
#define HANDLE_SINGLE_STEP_TRAP_fname_index 11
#define ENTER_ALIEN_CALLBACK_fname_index 12
#endif
