stop-the-world.o stop-the-world.d : stop-the-world.c genesis/thread.h lispobj.h \
 genesis/sbcl.h gencgc-alloc-region.h arch.h os.h genesis/sbcl.h \
 runtime.h lispobj.h align.h target-os.h target-arch-os.h target-arch.h \
 thread.h globals.h genesis/symbol.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/static-symbols.h \
 genesis/vector.h interrupt.h validate.h gc-typedefs.h gc-assert.h \
 interr.h gc.h immobile-space.h gencgc-impl.h gencgc-alloc-region.h \
 genesis/cardmarks.h genesis/cons.h genesis/weak-pointer.h hopscotch.h \
 code.h genesis/closure.h forwarding-ptr.inc pseudo-atomic.h lispregs.h \
 target-lispregs.h genesis/regnames.h atomiclog.inc
